package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.defs.JumpProfile
import ru.dbotthepony.kstarbound.defs.MovementParameters
import ru.dbotthepony.kstarbound.defs.player.ActorMovementModifiers
import ru.dbotthepony.kstarbound.util.GameTimer
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import kotlin.math.PI
import kotlin.math.absoluteValue
import kotlin.math.sign

abstract class AbstractActorMovementController : AbstractMovementController() {
	abstract var controlRun: Boolean
	abstract var controlCrouch: Boolean
	abstract var controlDown: Boolean
	abstract var lastControlDown: Boolean
	abstract var controlFly: Vector2d?
	abstract var controlFace: Direction?

	abstract var isRunning: Boolean
		protected set
	abstract var isWalking: Boolean
		protected set
	abstract var isCrouching: Boolean
		protected set
	abstract var isFlying: Boolean
		protected set
	abstract var isFalling: Boolean
		protected set
	abstract var isJumping: Boolean
		protected set

	abstract var canJump: Boolean
	abstract var controlJump: Boolean
	abstract var controlJumpAnyway: Boolean

	abstract var lastControlJump: Boolean
		protected set
	abstract var lastControlCrouch: Boolean
		protected set

	abstract var isGroundMovement: Boolean
		protected set
	abstract var isLiquidMovement: Boolean
		protected set

	abstract var controlRotationRate: Double
	abstract var controlAcceleration: Vector2d
	abstract var controlForce: Vector2d

	abstract var fallThroughSustain: Int
		protected set

	// Target horizontal velocity for walking / running
	abstract var targetHorizontalAmbulatingVelocity: Double
		protected set

	abstract var controlPathMove: Pair<Vector2d, Boolean>?
	abstract var pathMoveResult: Pair<Vector2d, Boolean>?

	abstract var controlMove: Direction?

	abstract var actorMovementParameters: ActorMovementParameters
	abstract var movementModifiers: ActorMovementModifiers

	abstract var controlActorMovementParameters: ActorMovementParameters
	abstract var controlMovementModifiers: ActorMovementModifiers

	abstract val approachVelocities: MutableList<ApproachVelocityCommand>
	abstract val approachVelocityAngles: MutableList<ApproachVelocityAngleCommand>

	abstract var movingDirection: Direction?
	abstract var facingDirection: Direction?

	// this is set internally on each move step
	final override var movementParameters: MovementParameters = MovementParameters.EMPTY

	abstract var anchorEntity: DynamicEntity?

	var pathController: PathController? = null
	var groundMovementSustainTimer: GameTimer = GameTimer(0.0)
	var reJumpTimer: GameTimer = GameTimer(0.0)
	var jumpHoldTimer: GameTimer? = null

	data class ApproachVelocityCommand(
		val target: Vector2d,
		val maxControlForce: Double
	)

	data class ApproachVelocityAngleCommand(
		val alongAngle: Double,
		val targetVelocity: Double,
		val maxControlForce: Double,
		val positiveOnly: Boolean
	)

	fun calculateMovementParameters(base: ActorMovementParameters): MovementParameters {
		val mass = base.mass
		val gravityMultiplier = base.gravityMultiplier

		val liquidBuoyancy = base.liquidBuoyancy
		val airBuoyancy = base.airBuoyancy
		val bounceFactor = base.bounceFactor
		val stopOnFirstBounce = base.stopOnFirstBounce
		val enableSurfaceSlopeCorrection = base.enableSurfaceSlopeCorrection
		val slopeSlidingFactor = base.slopeSlidingFactor
		val maxMovementPerStep = base.maxMovementPerStep

		val collisionPoly = if (isCrouching) base.crouchingPoly else base.standingPoly

		val stickyCollision = base.stickyCollision
		val stickyForce = base.stickyForce

		val airFriction = base.airFriction
		val liquidFriction = base.liquidFriction

		// If we are traveling in the correct direction while in a movement mode that
		// requires contact with the ground (ambulating i.e. walking or running), and
		// not traveling faster than our target horizontal movement, then apply the
		// special 'ambulatingGroundFriction'.
		val relativeXVelocity = velocity.x - surfaceVelocity.x
		val useAmbulatingGroundFriction = (isWalking || isRunning) &&
			targetHorizontalAmbulatingVelocity.sign == relativeXVelocity.sign &&
			relativeXVelocity.absoluteValue <= targetHorizontalAmbulatingVelocity.absoluteValue

		val groundFriction = if (useAmbulatingGroundFriction) base.ambulatingGroundFriction else base.normalGroundFriction

		val ignorePlatformCollision = fallThroughSustain > 0 || controlFly != null || controlDown

		val collisionEnabled = base.collisionEnabled
		val frictionEnabled = base.frictionEnabled
		val gravityEnabled = base.gravityEnabled

		val maximumPlatformCorrection = base.maximumPlatformCorrection
		val maximumPlatformCorrectionVelocityFactor = base.maximumPlatformCorrectionVelocityFactor

		val physicsEffectCategories = base.physicsEffectCategories

		val maximumCorrection = base.maximumCorrection
		val speedLimit = base.speedLimit

		return MovementParameters(
			mass = mass,
			gravityMultiplier = gravityMultiplier,
			liquidBuoyancy = liquidBuoyancy,
			airBuoyancy = airBuoyancy,
			bounceFactor = bounceFactor,
			stopOnFirstBounce = stopOnFirstBounce,
			enableSurfaceSlopeCorrection = enableSurfaceSlopeCorrection,
			slopeSlidingFactor = slopeSlidingFactor,
			maxMovementPerStep = maxMovementPerStep,
			maximumCorrection = maximumCorrection,
			speedLimit = speedLimit,
			stickyCollision = stickyCollision,
			stickyForce = stickyForce,
			airFriction = airFriction,
			liquidFriction = liquidFriction,
			groundFriction = groundFriction,
			collisionEnabled = collisionEnabled,
			frictionEnabled = frictionEnabled,
			gravityEnabled = gravityEnabled,
			maximumPlatformCorrection = maximumPlatformCorrection,
			maximumPlatformCorrectionVelocityFactor = maximumPlatformCorrectionVelocityFactor,
			physicsEffectCategories = physicsEffectCategories,
			ignorePlatformCollision = ignorePlatformCollision,
			collisionPoly = collisionPoly,
		)
	}

	fun updateMovementParameters(base: ActorMovementParameters): MovementParameters {
		val params = calculateMovementParameters(base)
		movementParameters = params
		return params
	}

	open fun clearControls() {
		controlRotationRate = 0.0
		controlAcceleration = Vector2d.ZERO
		controlForce = Vector2d.ZERO
		controlRun = false
		controlCrouch = false
		controlJump = false
		controlJumpAnyway = false
		controlFly = null
		controlPathMove = null
		controlActorMovementParameters = ActorMovementParameters.EMPTY
		controlMovementModifiers = ActorMovementModifiers.EMPTY
	}

	override fun move() {
		// TODO: anchor entity

		if (anchorEntity?.isSpawned != true)
			anchorEntity = null

		val anchorEntity = anchorEntity

		if (anchorEntity != null) {
			controlRun = false
			controlCrouch = false
			isFlying = false
			isFalling = false
			canJump = false
			controlJump = false
			isGroundMovement = false
			isLiquidMovement = false

			velocity = (anchorEntity.position - position) / Starbound.TICK_TIME_ADVANCE
			super.move()
			position = anchorEntity.position
		} else {
			val movementParameters = actorMovementParameters.merge(controlActorMovementParameters)
			val movementModifiers = movementModifiers.combine(controlMovementModifiers)

			if (movementModifiers.movementSuppressed) {
				controlMove = null
				controlRun = false
				controlCrouch = false
				controlDown = false
				controlJump = false
				controlPathMove = null
			}

			// controlling any other movement overrides the pathing
			if (controlMove != null || controlCrouch || controlDown || controlJump || controlFly != null || approachVelocities.isNotEmpty() || approachVelocityAngles.isNotEmpty()) {
				controlPathMove = null
			}

			if (controlPathMove != null && pathMoveResult == null) {
				if (appliedForceRegion) {
					pathController?.reset()
				} else if (!pathController!!.isPathfinding) {
					// TODO: path move code
				} else {

				}
			} else {
				pathController = null
			}

			if (controlFly != null)
				controlMove = null

			if ((controlDown && !lastControlDown) || controlFly != null)
				fallThroughSustain = movementParameters.fallThroughSustainFrames ?: 0
			else if (fallThroughSustain > 0)
				fallThroughSustain--

			updateMovementParameters(movementParameters)

			targetHorizontalAmbulatingVelocity = 0.0

			rotation = (rotation + controlRotationRate * Starbound.TICK_TIME_ADVANCE) % (PI * 2.0)
			velocity += controlAcceleration * Starbound.TICK_TIME_ADVANCE + controlForce / mass * Starbound.TICK_TIME_ADVANCE

			approachVelocities.forEach {
				approachVelocity(it.target, it.maxControlForce)
			}

			approachVelocityAngles.forEach {
				approachVelocityAlongAngle(it.alongAngle, it.targetVelocity, it.maxControlForce, it.positiveOnly)
			}

			isLiquidMovement = liquidPercentage >= (actorMovementParameters.minimumLiquidPercentage ?: 0.0)
			val liquidImpedance = liquidPercentage * (actorMovementParameters.liquidImpedance ?: 0.0)

			var updatedMovingDirection: Direction? = null
			val isRunning = controlRun && !movementModifiers.runningSuppressed

			if (controlFly != null) {
				var flyVelocity = controlFly!!

				if (flyVelocity.lengthSquared != 0.0)
					flyVelocity = flyVelocity.unitVector * (actorMovementParameters.flySpeed ?: 0.0)

				if (isLiquidMovement)
					approachVelocity(flyVelocity * (1.0 - liquidImpedance) * movementModifiers.speedModifier, (movementParameters.liquidForce ?: 0.0) * movementModifiers.liquidMovementModifier)
				else
					approachVelocity(flyVelocity * movementModifiers.speedModifier, movementParameters.airForce ?: 0.0)

				if (flyVelocity.x > 0.0)
					updatedMovingDirection = Direction.RIGHT
				else if (flyVelocity.x < 0.0)
					updatedMovingDirection = Direction.LEFT

				groundMovementSustainTimer = GameTimer(0.0)
			} else {
				val jumpModifier: Double
				val jumpProfile: JumpProfile

				if (isLiquidMovement) {
					jumpModifier = movementModifiers.liquidJumpModifier
					jumpProfile = movementParameters.liquidJumpProfile.copy(jumpSpeed = (movementParameters.liquidJumpProfile.jumpSpeed ?: 0.0) * (1.0 - liquidImpedance))
				} else {
					jumpModifier = movementModifiers.airJumpModifier
					jumpProfile = movementParameters.airJumpProfile
				}

				var startJump = false
				var holdJump = false

				// If we are on the ground, then reset the ground movement sustain timer
				// to the maximum.  If we are not on the ground or near the ground
				// according to the nearGroundCheckDistance, and we are past the minimum
				// sustain time, then go ahead and immediately clear the ground movement
				// sustain timer.
				val minGroundSustain = movementParameters.groundMovementMinimumSustain ?: 0.0
				val maxGroundSustain = movementParameters.groundMovementMaximumSustain ?: 0.0
				val groundCheckDistance = movementParameters.groundMovementCheckDistance ?: 0.0

				groundMovementSustainTimer.tick()

				if (isOnGround) {
					groundMovementSustainTimer = GameTimer(maxGroundSustain)
				} else if (!groundMovementSustainTimer.hasFinished && groundCheckDistance > 0.0 && maxGroundSustain - groundMovementSustainTimer.timer > minGroundSustain) {
					val collideAny = localHitboxes
						.map { it + Vector2d(0.0, -groundCheckDistance) }
						.anyMatch {
							world.polyIntersects(it, { it.type >= CollisionType.PLATFORM })
						}

					if (collideAny)
						groundMovementSustainTimer = GameTimer(0.0)
				}

				val standingJumpable = !groundMovementSustainTimer.hasFinished
				val controlJump = this.controlJump && (!movementModifiers.jumpingSuppressed || controlJumpAnyway)

				// We are doing a jump if m_reJumpTimer has run out and there has been a
				// new m_controlJump command which was just recently triggered.  If
				// jumpProfile.autoJump is set, then we don't care whether it is a new
				// m_controlJump command, m_controlJump can be held.
				if (reJumpTimer.hasFinished && controlJump && (jumpProfile.autoJump == true || !lastControlJump)) {
					if (standingJumpable || jumpProfile.multiJump == true || controlJumpAnyway)
						startJump = true
				} else if (isJumping && controlJump && jumpHoldTimer?.hasFinished != true) {
					if (jumpProfile.collisionCancelled != true || collisionCorrection.y >= 0.0)
						holdJump = true
				}

				if (startJump) {
					isJumping = true
					reJumpTimer = GameTimer(jumpProfile.reJumpDelay ?: 0.0)

					jumpHoldTimer = if ((jumpProfile.jumpHoldTime ?: 0.0) >= 0.0) {
						GameTimer(jumpProfile.jumpHoldTime ?: 0.0)
					} else {
						null
					}

					velocity = velocity.copy(y = velocity.y + (jumpProfile.jumpSpeed ?: 0.0) * (jumpProfile.jumpInitialPercentage ?: 0.0) * jumpModifier)
					groundMovementSustainTimer = GameTimer(0.0)
				} else if (holdJump) {
					reJumpTimer.tick()
					jumpHoldTimer?.tick()

					approachYVelocity((jumpProfile.jumpSpeed ?: 0.0) * jumpModifier, (jumpProfile.jumpControlForce ?: 0.0) * jumpModifier)
				} else {
					isJumping = false
					reJumpTimer.tick()
				}

				if (controlMove == Direction.LEFT) {
					updatedMovingDirection = Direction.LEFT
					targetHorizontalAmbulatingVelocity = -1.0 * (if (isRunning) movementParameters.runSpeed ?: 0.0 else movementParameters.walkSpeed ?: 0.0) * movementModifiers.speedModifier
				} else if (controlMove == Direction.RIGHT) {
					updatedMovingDirection = Direction.RIGHT
					targetHorizontalAmbulatingVelocity = 1.0 * (if (isRunning) movementParameters.runSpeed ?: 0.0 else movementParameters.walkSpeed ?: 0.0) * movementModifiers.speedModifier
				}

				if (isLiquidMovement)
					targetHorizontalAmbulatingVelocity *= 1.0 - liquidImpedance

				// don't ambulate if we're already moving faster than the target velocity in the direction of ambulation
				val ambulationWouldAccelerate = (targetHorizontalAmbulatingVelocity + surfaceVelocity.x).absoluteValue > velocity.x.absoluteValue ||
					targetHorizontalAmbulatingVelocity < 0 != velocity.x < 0

				if (targetHorizontalAmbulatingVelocity != 0.0 && ambulationWouldAccelerate) {
					val ambulatingAccelerate = if (isOnGround)
						(movementParameters.groundForce ?: 0.0) * movementModifiers.groundMovementModifier
					else if (isLiquidMovement)
						(movementParameters.liquidForce ?: 0.0) * movementModifiers.liquidMovementModifier
					else
						movementParameters.airForce ?: 0.0

					approachXVelocity(targetHorizontalAmbulatingVelocity + surfaceVelocity.x, ambulatingAccelerate)
				}
			}

			movingDirection = updatedMovingDirection

			if (!movementModifiers.facingSuppressed) {
				if (controlFace != null)
					facingDirection = controlFace
				else if (updatedMovingDirection != null)
					facingDirection = updatedMovingDirection
				else if (controlPathMove != null && pathController?.controlFace != null)
					facingDirection = pathController?.controlFace
			}

			isGroundMovement = !groundMovementSustainTimer.hasFinished

			if (isGroundMovement) {
				this.isRunning = isRunning && controlMove != null
				this.isWalking = !isRunning && controlMove != null
				this.isCrouching = controlCrouch && controlMove != null
			}

			isFlying = controlFly != null
			isFalling = (velocity.y < (movementParameters.fallStatusSpeedMin ?: 0.0)) && !isGroundMovement

			super.move()

			lastControlDown = controlDown
			lastControlJump = controlJump

			if (isLiquidMovement)
				canJump = reJumpTimer.hasFinished && (!groundMovementSustainTimer.hasFinished || movementParameters.liquidJumpProfile.multiJump == true)
			else
				canJump = reJumpTimer.hasFinished && (!groundMovementSustainTimer.hasFinished || movementParameters.airJumpProfile.multiJump == true)
		}

		clearControls()
	}
}
