package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.util.AABB
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.render.LayeredRenderer
import ru.dbotthepony.kstarbound.client.render.RenderLayer
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.World

/**
 * Entities with dynamics (Player, Drops, Projectiles, NPCs, etc)
 */
abstract class DynamicEntity(path: String) : AbstractEntity(path) {
	private var forceChunkRepos = false

	var position = Vector2d()
		set(value) {
			val old = field

			if (isSpawned) {
				field = world.geometry.wrap(value)

				val oldChunkPos = world.geometry.chunkFromCell(old)
				val newChunkPos = world.geometry.chunkFromCell(field)

				chunkPos = newChunkPos

				if (oldChunkPos != newChunkPos || forceChunkRepos) {
					chunk = world.chunkMap[newChunkPos]
					forceChunkRepos = false
				}
			} else {
				field = value
			}
		}

	abstract val movement: AbstractMovementController
	final override var chunkPos: ChunkPos = ChunkPos.ZERO
		private set

	override fun onSpawn(world: World<*, *>) {
		world.dynamicEntities.add(this)
		forceChunkRepos = true
		position = position
	}

	override fun onRemove(world: World<*, *>) {
		world.dynamicEntities.remove(this)
	}

	override fun render(client: StarboundClient, layers: LayeredRenderer) {
		layers.add(RenderLayer.Overlay.point()) {
			val hitboxes = movement.localHitboxes.toList()
			if (hitboxes.isEmpty()) return@add

			hitboxes.forEach { it.render(client) }

			world.queryCollisions(
				hitboxes.stream().map { it.aabb }.reduce(AABB::combine).get().enlarge(2.0, 2.0)
			).filter(movement::shouldCollideWithBody).forEach { it.poly.render(client, BLOCK_COLLISION_COLOR) }
		}
	}

	companion object {
		val BLOCK_COLLISION_COLOR = RGBAColor(65, 179, 217)
		const val SEPARATION_STEPS = 3
		const val SEPARATION_TOLERANCE = 0.001
	}
}
